//ModifyInput.cpp : A simple console app which reads data from a 16 bit mono wav file and modifies its frequency data before playing it out

#include "stdafx.h"
#include "..\CMMSystem.h"
#include "Fourier.h" //If you get a compiler error about this missing header file, please review the CMMSystem.htm documentation
#include <climits>
#include <vector>


#pragma warning(suppress: 26429 26485 26487)
int _tmain(int argc, TCHAR* argv[], TCHAR* /*envp*/[]) //NOLINT(modernize-avoid-c-arrays)
{
  //parse the command line
  if (argc != 2)
  {
    _ftprintf(stderr, _T("ModifyInput InputWavFile\n"));
    return 2;
  }

  try
  {
    //try and open the wav file we are going to play
    CMMIOFile file;
    try
    {
#pragma warning(suppress: 26481 26486)
      file.Open(argv[1], MMIO_READ | MMIO_DENYWRITE | MMIO_ALLOCBUF);
    }
#pragma warning(suppress: 26496)
    catch(CMMIOException& e)
    {
#pragma warning(suppress: 26481)
      _ftprintf(stderr, _T("Failed to open file %s, Error:%u\n"), argv[1], e.m_Error);
      return 3;
    }

    //Check that it is a WAV RIFF file
    MMCKINFO ckInRIFF{};
    ckInRIFF.fccType = mmioFOURCC('W', 'A', 'V', 'E');
    try
    {
      file.Descend(&ckInRIFF, nullptr, MMIO_FINDRIFF);
    }
#pragma warning(suppress: 26496)
    catch(CMMIOException& e)
    {
#pragma warning(suppress: 26481)
      _ftprintf(stderr, _T("File %s does not appear to be a properly formatted WAV file, Error:%u\n"), argv[1], e.m_Error);
      return 4;
    }

    //Find the "FMT" chunk (form type "FMT"); it must be 
    //a subchunk of the "RIFF" chunk.
    MMCKINFO cki{};
    cki.ckid = mmioFOURCC('f', 'm', 't', ' ');
    try
    {
      file.Descend(&cki, &ckInRIFF, MMIO_FINDCHUNK);

      if (cki.cksize < sizeof(PCMWAVEFORMAT))
      {
#pragma warning(suppress: 26481 26486)
        _ftprintf(stderr, _T("File %s does not appear to be contain a proper \"FMT\" WAV file chunk\n"), argv[1]);
        return 5;
      }
    }
#pragma warning(suppress: 26496)
    catch(CMMIOException& e)
    {
#pragma warning(suppress: 26481)
      _ftprintf(stderr, _T("File %s does not appear to be contain a proper \"FMT\" WAV file chunk, Error:%u\n"), argv[1], e.m_Error);
      return 5;
    }

    //Read in the FMT chunk (which in the case of a WAV file is a PCMWAVEFORMAT struct)
    PCMWAVEFORMAT pcm{};
    file.Read(&pcm, sizeof(pcm));
    if ((pcm.wBitsPerSample != 16) || (pcm.wf.nChannels != 1))
    {
#pragma warning(suppress: 26481 26486)
      _ftprintf(stderr, _T("File %s does not appear to be a mono 16 bit wav file"), argv[1]);
      return 6;
    }

    //Ascend out of the "FMT" subchunk, so that we can descend to find the data next
    file.Ascend(&cki); 

    //Find the data subchunk. The current file position should be at 
    //the beginning of the data chunk; however, you should not make 
    //this assumption. Use mmioDescend to locate the data chunk. 
    cki.ckid = mmioFOURCC('d', 'a', 't', 'a');
    try
    {
      file.Descend(&cki, &ckInRIFF, MMIO_FINDCHUNK);
    }
#pragma warning(suppress: 26496)
    catch(CMMIOException& e)
    {
#pragma warning(suppress: 26481)
      _ftprintf(stderr, _T("File %s does not appear to contain any audio data, Error:%u\n"), argv[1], e.m_Error);
      return 7;
    }

    //Open the wave input device
    WAVEFORMATEX wfe{};
    wfe.wFormatTag = pcm.wf.wFormatTag;
    wfe.nChannels = pcm.wf.nChannels;
    wfe.nSamplesPerSec = pcm.wf.nSamplesPerSec;
    wfe.wBitsPerSample = pcm.wBitsPerSample;
#pragma warning(suppress: 26472)
    wfe.nBlockAlign = static_cast<WORD>((wfe.nChannels * wfe.wBitsPerSample) >> 3);
    wfe.nAvgBytesPerSec = wfe.nSamplesPerSec * wfe.nBlockAlign;
    wfe.cbSize = 0;

    //Also open up the output device
    CWaveOut writer;

    //Also it to allocate its buffers
    writer.CreateBuffers();

    //Open the output device using the same settings as the input device
    writer.Open(&wfe);

    //Allocate memory for the various arrays we will require
    std::vector<double> RealIn{1024 , std::allocator<double>{}};
    std::vector<short> inbuffer{1024 , std::allocator<short>{}};
    std::vector<double> RealOut{1024 , std::allocator<double>{}};
    std::vector<double> ImagOut{1024 , std::allocator<double>{}};
    std::vector<double> ModRealOut{1024 , std::allocator<double>{}};
    std::vector<double> ModImagOut{1024 , std::allocator<double>{}};
    std::vector<short> outbuffer{1024 , std::allocator<short>{}};

    //the main record loop, basically read from the input
    //and write it the wave file, until a key is hit. What could be simpler!
    DWORD dwRead{0};
    while (dwRead < cki.cksize)
    {
      //Read from the file in chunks of 1K
#pragma warning(suppress: 26486)
      const LONG nReadBytes{file.Read(inbuffer.data(), sizeof(short) * 1024)};

      if (nReadBytes)
      {
        dwRead += nReadBytes;

        //Perform the FFT on the input data
        for (int i{0}; i<1024; i++)
#pragma warning(suppress: 26446 26481 26482)
          RealIn[i] = inbuffer[i];
#pragma warning(suppress: 26486)
        fft_double(1024, false, RealIn.data(), nullptr, RealOut.data(), ImagOut.data());

        //Shift up all the frequencies up somewhat
        constexpr const int nShift = 8;
        for (intptr_t i{511}; i>=0; i--)
        {
          if (i > nShift)
          {
#pragma warning(suppress: 26446 26481 26482 26489)
            RealOut[i] = RealOut[i - nShift];
#pragma warning(suppress: 26446 26481 26482 26489)
            ImagOut[i] = ImagOut[i - nShift];

#pragma warning(suppress: 26446 26481 26482 26489)
            RealOut[1024 - i] = RealOut[i];
#pragma warning(suppress: 26446 26481 26482 26489)
            ImagOut[1024 - i] = ImagOut[i];
          }
          else if (i <= nShift)
          {
#pragma warning(suppress: 26446 26481 26482 26489)
            RealOut[i] = 0;
#pragma warning(suppress: 26446 26481 26482 26489)
            ImagOut[i] = 0;

#pragma warning(suppress: 26446 26481 26482 26489)
            RealOut[1023 - i] = 0;
#pragma warning(suppress: 26446 26481 26482 26489)
            ImagOut[1023 - i] = 0;
          }
        }
        //Attenuate some of the higher frequencies after we have shifted them up
        for (intptr_t i{90}; i<512; i++)
        {
#pragma warning(suppress: 26446 26481 26482 26489)
          RealOut[i] = RealOut[i] * 0.5;
#pragma warning(suppress: 26446 26481 26482 26489)
          ImagOut[i] = ImagOut[i] * 0.5;
#pragma warning(suppress: 26446 26481 26482 26489)
          RealOut[1023 - i] = RealOut[i];
#pragma warning(suppress: 26446 26481 26482 26489)
          ImagOut[1023 - i] = ImagOut[i];
        }

        //Perform the inverse FFT
#pragma warning(suppress: 26486)
        fft_double(1024, true, RealOut.data(), ImagOut.data(), ModRealOut.data(), ModImagOut.data());

        //Convert back to a short array
        for (int i{0}; i<1024; i++)
        {
#pragma warning(suppress: 26446 26481 26482 26489)
          if (ModRealOut[i] < 0)
#pragma warning(suppress: 26446 26481 26482 26489)
            outbuffer[i] = static_cast<short>(ModRealOut[i] - 0.5);
          else
#pragma warning(suppress: 26446 26481 26482 26489)
            outbuffer[i] = static_cast<short>(ModRealOut[i] + 0.5);
        }

        //* //If you want to use the unaltered data
        //for (int i{0}; i<1024; i++)
        //  outbuffer[i] = inbuffer[i];


        //Finally, write out the modified data to the wave out device
#pragma warning(suppress: 26486)
        writer.Write(outbuffer.data(), sizeof(short)*1024);
      }
    }
  }
#pragma warning(suppress: 26496)
  catch(CMMIOException& e)
  {
    _ftprintf(stderr, _T("An unexpected mmio exception occurred, Error:%u\n"), e.m_Error);
  }
#pragma warning(suppress: 26496)
  catch(CWaveOutException& e)
  {
    _ftprintf(stderr, _T("An unexpected waveOut exception occurred, Error:%u\n"), e.m_Error);
  }

  return 0;
}
