//PlayWav.cpp : A simple console app which plays a wav file using the CMMIOFile and CWaveOut wrapper classes

#include "stdafx.h"
#include "..\CMMSystem.h"
#include <array>


#pragma warning(suppress: 26429 26485 26487)
int _tmain(int argc, TCHAR* argv[], TCHAR* /*envp*/[]) //NOLINT(modernize-avoid-c-arrays)
{
  //parse the command line
  if (argc != 2)
  {
    _ftprintf(stderr, _T("PlayWav WavFile\n"));
    return 2;
  }

/* //Some test code to try out the mixer APIs
  try
  {
    CMixer mixer;
    UINT nMixers{CMixer::GetNumDevs()};
    for (UINT i{0}; i<nMixers; i++)
    {
      MIXERCAPS mxc;
      CMixer::GetDevCaps(i, &mxc);
    }

    //Open the first mixer
    mixer.Open(0);

    MIXERCAPS mxc;
    mixer.GetDevCaps(&mxc);

    MIXERCONTROLDETAILS mxcd{};
    mxcd.cbStruct = sizeof(mxcd);
    mxcd.dwControlID = 1;
    //mixer.GetControlDetails(&mxcd, 0);
    //mixer.SetControlDetails(&mxcd, 0);

    MIXERLINE mxl{};
    mxl.cbStruct = sizeof(MIXERLINE);
    mxl.dwComponentType = MIXERLINE_COMPONENTTYPE_SRC_COMPACTDISC;
    mixer.GetLineInfo(&mxl, MIXER_GETLINEINFOF_COMPONENTTYPE);

    UINT uMxId{0};
    mixer.GetID(&uMxId);

    MIXERLINECONTROLS mxlc{};
    mxlc.cbStruct = sizeof(MIXERLINECONTROLS);
    mxlc.cbmxctrl  = sizeof(MIXERCONTROL);
    mxlc.cControls = 1;
    MIXERCONTROL mc{};
    mc.cbStruct = sizeof(MIXERCONTROL);
    mxlc.pamxctrl = &mc;
    mixer.GetLineControls(&mxlc, MIXER_GETLINECONTROLSF_ONEBYID);

    mixer.Close();
  }
  catch(CMixerException& e)
  {
    _ftprintf(stderr, _T("An unexpected mixer exception occurred, Error:%u\n"), e.m_Error);
  }
*/

  try
  {
    //Create an instance of the class which does the audio output
    CWaveOut writer;

    //Ask it to allocate its buffers
    writer.CreateBuffers();

    //try and open the wav file we are going to play
    CMMIOFile file;
    try
    {
#pragma warning(suppress: 26481 26486)
      file.Open(argv[1], MMIO_READ | MMIO_DENYWRITE | MMIO_ALLOCBUF);
    }
#pragma warning(suppress: 26496)
    catch(CMMIOException& e)
    {
#pragma warning(suppress: 26481)
      _ftprintf(stderr, _T("Failed to open file %s, Error:%u\n"), argv[1], e.m_Error);
      return 3;
    }

    //Check that it is a WAV RIFF file
    MMCKINFO ckInRIFF{};
    ckInRIFF.fccType = mmioFOURCC('W', 'A', 'V', 'E');
    try
    {
      file.Descend(&ckInRIFF, nullptr, MMIO_FINDRIFF);
    }
#pragma warning(suppress: 26496)
    catch(CMMIOException& e)
    {
#pragma warning(suppress: 26481)
      _ftprintf(stderr, _T("File %s does not appear to be a properly formatted WAV file, Error:%u\n"), argv[1], e.m_Error);
      return 4;
    }

    //Find the "FMT" chunk (form type "FMT"); it must be 
    //a subchunk of the "RIFF" chunk. 
    MMCKINFO cki{};
    cki.ckid = mmioFOURCC('f', 'm', 't', ' ');
    try
    {
      file.Descend(&cki, &ckInRIFF, MMIO_FINDCHUNK);

      if (cki.cksize < sizeof(PCMWAVEFORMAT))
      {
#pragma warning(suppress: 26481 26486)
        _ftprintf(stderr, _T("File %s does not appear to be contain a proper \"FMT\" WAV file chunk\n"), argv[1]);
        return 5;
      }
    }
#pragma warning(suppress: 26496)
    catch(CMMIOException& e)
    {
#pragma warning(suppress: 26481)
      _ftprintf(stderr, _T("File %s does not appear to be contain a proper \"FMT\" WAV file chunk, Error:%u\n"), argv[1], e.m_Error);
      return 5;
    }

    //Read in the FMT chunk (which in the case of a WAV file is a PCMWAVEFORMAT struct
    PCMWAVEFORMAT pcm{};
    file.Read(&pcm, sizeof(pcm));

    //Ascend out of the "FMT" subchunk, so that we can descend to find the data next
    file.Ascend(&cki); 

    //Find the data subchunk. The current file position should be at 
    //the beginning of the data chunk; however, you should not make 
    //this assumption. Use mmioDescend to locate the data chunk. 
    cki.ckid = mmioFOURCC('d', 'a', 't', 'a');
    try
    {
      file.Descend(&cki, &ckInRIFF, MMIO_FINDCHUNK);
    }
#pragma warning(suppress: 26496)
    catch(CMMIOException& e)
    {
#pragma warning(suppress: 26481)
      _ftprintf(stderr, _T("File %s does not appear to contain any audio data, Error:%u\n"), argv[1], e.m_Error);
      return 6;
    }

    //Open the wave output device
    WAVEFORMATEX wfe{};
    wfe.wFormatTag = pcm.wf.wFormatTag;
    wfe.nChannels = pcm.wf.nChannels;
    wfe.nSamplesPerSec = pcm.wf.nSamplesPerSec;
    wfe.wBitsPerSample = pcm.wBitsPerSample;
    wfe.nBlockAlign = pcm.wf.nBlockAlign;
    wfe.nAvgBytesPerSec = pcm.wf.nAvgBytesPerSec;
    wfe.cbSize = 0;
    writer.Open(&wfe);

#ifdef _DEBUG
    WAVEOUTCAPS woc{};
    writer.GetDevCaps(&woc);
    writer.GetNumDevs();
    writer.BreakLoop();
    const DWORD dwRate{writer.GetPlaybackRate()};
    writer.SetPlaybackRate(dwRate);
    MMTIME mmt;
    mmt.wType = TIME_MS;
    writer.GetPosition(&mmt, sizeof(mmt));
    const DWORD dwVolume{writer.GetVolume()};
    writer.SetVolume(dwVolume);
    writer.Pause();
    writer.Reset();
    writer.Restart();
#endif //#ifdef _DEBUG

    //the main playback loop, basically read from the wave file
    //and write it the wave output class. What could be simpler!
    DWORD dwRead{0};
    while ((dwRead < cki.cksize))
    {
      //Read from the file in chunks of 1K
      std::array<BYTE, 1024> buffer{};
#pragma warning(suppress: 26472)
      const LONG nReadBytes{file.Read(buffer.data(), static_cast<LONG>(buffer.size()))};

      if (nReadBytes)
      {
        //Increment the current read pos
        dwRead += nReadBytes;

        //Write the data to the output
#pragma warning(suppress: 26485)
        writer.Write(buffer.data(), nReadBytes);
      }
    }
  }
#pragma warning(suppress: 26496)
  catch(CMMIOException& e)
  {
    _ftprintf(stderr, _T("An unexpected MMIO exception occurred, Error:%u\n"), e.m_Error);
  }
#pragma warning(suppress: 26496)
  catch(CWaveOutException& e)
  {
    _ftprintf(stderr, _T("An unexpected waveOut exception occurred, Error:%u\n"), e.m_Error);
  }

  return 0;
}
