//RecordWav.cpp : A simple console app which records a wav file using the CMMIOFile and CWaveIn wrapper classes

#include "stdafx.h"
#include "..\CMMSystem.h"
#include <array>

#pragma warning(suppress: 26429 26485 26487)
int _tmain(int argc, TCHAR* argv[], TCHAR* /*envp*/[]) //NOLINT(modernize-avoid-c-arrays)
{
  //parse the command line
  if (argc != 5)
  {
    _ftprintf(stderr, _T("RecordWav WavFile Channels SamplesPerSecond BitsPerSample\n"));
    return 2;
  }

#pragma warning(suppress: 26472 26481 26486)
  const auto wChannels{static_cast<WORD>(_ttoi(argv[2]))};
#pragma warning(suppress: 26472 26481 26486)
  const auto dwSamplesPerSecond{static_cast<DWORD>(_ttoi(argv[3]))};
#pragma warning(suppress: 26472 26481 26486)
  const auto wBitsPerSample{static_cast<WORD>(_ttoi(argv[4]))};

  try
  {
    _tprintf(_T("Press any key to stop recording\n"));

    //Create an instance of the class which does the audio input
    CWaveIn reader;

    //Ask it to allocate its buffers
    reader.CreateBuffers();

    //try and open the wav file we are going to write to
    CMMIOFile file;
    try
    {
#pragma warning(suppress: 26481 26486)
      file.Open(argv[1], MMIO_WRITE | MMIO_CREATE | MMIO_DENYWRITE | MMIO_ALLOCBUF);
    }
#pragma warning(suppress: 26496)
    catch(CMMIOException& e)
    {
#pragma warning(suppress: 26481)
      _ftprintf(stderr, _T("Failed to open file %s, Error:%u\n"), argv[1], e.m_Error);
      return 3;
    }

    //Create the wave chunk
    MMCKINFO ckOutRIFF{};
    ckOutRIFF.fccType = mmioFOURCC('W', 'A', 'V', 'E');
    file.CreateChunk(&ckOutRIFF, MMIO_CREATERIFF);

    //Create the "FMT" chunk as a subchunk of the "RIFF" chunk. 
    MMCKINFO ckOut{};
    ckOut.ckid = mmioFOURCC('f', 'm', 't', ' ');
    ckOut.cksize = sizeof(PCMWAVEFORMAT);
    file.CreateChunk(&ckOut, 0);

    //Write out the FMT chunk
    PCMWAVEFORMAT wfe{};
    wfe.wf.wFormatTag = WAVE_FORMAT_PCM;
    wfe.wf.nChannels = wChannels;
    wfe.wf.nSamplesPerSec = dwSamplesPerSecond;
    wfe.wBitsPerSample = wBitsPerSample;
#pragma warning(suppress: 26472)
    wfe.wf.nBlockAlign = static_cast<WORD>((wfe.wf.nChannels * wfe.wBitsPerSample) >> 3);
    wfe.wf.nAvgBytesPerSec = wfe.wf.nSamplesPerSec * wfe.wf.nBlockAlign;
    file.Write(&wfe, sizeof(wfe));

    //Ascend out of the "FMT" subchunk, so that we can descend to find the data next
    file.Ascend(&ckOut);

    //Create the data subchunk.
    ckOut.ckid = mmioFOURCC('d', 'a', 't', 'a');
    file.CreateChunk(&ckOut, 0);

    //Open the wave input device
    WAVEFORMATEX wfe2{};
    wfe2.wFormatTag = wfe.wf.wFormatTag;
    wfe2.nChannels = wfe.wf.nChannels;
    wfe2.nSamplesPerSec = wfe.wf.nSamplesPerSec;
    wfe2.wBitsPerSample = wfe.wBitsPerSample;
    wfe2.nBlockAlign = wfe.wf.nBlockAlign;
    wfe2.nAvgBytesPerSec = wfe.wf.nAvgBytesPerSec;
    wfe2.cbSize = 0;
    reader.Open(&wfe2);

    WAVEINCAPS wic;
    reader.GetDevCaps(&wic);

    //Start input on the device
    reader.Start();

    //the main record loop, basically read from the input
    //and write the data to the wave file, until a key is hit. What could be simpler!
    while (!_kbhit())
    {
      //Read from the file in chunks of 1K
      std::array<BYTE, 1024> buffer{};
#pragma warning(suppress: 26472)
      reader.Read(buffer.data(), static_cast<DWORD>(buffer.size()));

      //Write the data to file
#pragma warning(suppress: 26472)
      file.Write(buffer.data(), static_cast<DWORD>(buffer.size()));
    }

    //Ascend out of the data chunk, this will cause the chunk size of the
    //data chunk to be written
    file.Ascend(&ckOut);

    //Asecond out of the RIFF chunk, this will cause the chunk size of the
    //RIFF chunk to be written
    file.Ascend(&ckOutRIFF);

    reader.Stop();
  }
#pragma warning(suppress: 26496)
  catch(CMMIOException& e)
  {
    _ftprintf(stderr, _T("An unexpected MMIO exception occurred, Error:%u\n"), e.m_Error);
  }
#pragma warning(suppress: 26496)
  catch(CWaveInException& e)
  {
    _ftprintf(stderr, _T("An unexpected waveIn exception occurred, Error:%u\n"), e.m_Error);
  }

  return 0;
}
