#include "stdafx.h"
#include "ShowInput.h"
#include "ShowInputDoc.h"
#include "ShowInputView.h"
#include "..\CMMSystem.h"
#include <cmath>

#ifdef CSHOWINPUT_USECUDA
#pragma comment(lib, "cudart.lib")
#pragma comment(lib, "cufft.lib")
#else
#include "..\ModifyInput\Fourier.h" //If you get a compiler error about this missing header file, please review the CMMSystem.htm documentation
#endif


#ifdef _DEBUG
#define new DEBUG_NEW
#endif


#pragma warning(suppress: 26409 26433 26440 26477)
IMPLEMENT_DYNCREATE(CShowInputView, CView)

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CShowInputView, CView) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_DESTROY()
#ifndef CSHOWINPUT_NOD2D
  ON_REGISTERED_MESSAGE(AFX_WM_DRAW2D, &CShowInputView::OnDraw2D)
  ON_REGISTERED_MESSAGE(AFX_WM_RECREATED2DRESOURCES, &CShowInputView::OnRecreatedResources)
#endif //#ifndef COSMCTRL_NOD2D
  ON_WM_ERASEBKGND()
END_MESSAGE_MAP()

#pragma warning(suppress: 26455)
CShowInputView::CShowInputView() : m_pThread{nullptr},
                                   m_bWantStop{FALSE}
{
#pragma warning(suppress: 26485)
  memset(&m_inbuffer, 0, sizeof(m_inbuffer));

#ifndef CSHOWINPUT_NOD2D
  EnableD2DSupport();
#endif //#ifndef CSHOWINPUT_NOD2D

#ifdef CSHOWINPUT_USECUDA
  cudaError e{cudaMalloc(&m_InDevice, sizeof(cufftComplex) * _countof(m_inbuffer))};
  VERIFY(e == cudaSuccess);
  e = cudaMalloc(&m_OutDevice, sizeof(cufftComplex) * _countof(m_inbuffer));
  VERIFY(e == cudaSuccess);
  const cufftResult r{cufftPlan1d(&m_FFTPlan, _countof(m_inbuffer), CUFFT_C2C, 1)};
  VERIFY(r == CUFFT_SUCCESS);
  for (int i{0}; i<_countof(m_inbuffer); i++)
  {
#pragma warning(suppress: 26446 26482)
    m_OutHost[i].x = 0.0;
#pragma warning(suppress: 26446 26482)
    m_OutHost[i].y = 0.0;
#pragma warning(suppress: 26446 26482)
    m_InHost[i].x = 0.0;
#pragma warning(suppress: 26446 26482)
    m_InHost[i].y = 0.0;
  }
#else
  for (size_t i{0}; i<_countof(m_inbuffer); i++)
  {
#pragma warning(suppress: 26446 26482)
    m_ImagIn[i] = 0;
#pragma warning(suppress: 26446 26482)
    m_RealOut[i] = 0;
#pragma warning(suppress: 26446 26482)
    m_ImagOut[i] = 0;
  }
#endif
}

CShowInputView::~CShowInputView() //NOLINT(modernize-use-equals-default)
{
#ifdef CSHOWINPUT_USECUDA
  cufftDestroy(m_FFTPlan);
  cudaFree(m_InDevice);
  cudaFree(m_OutDevice);
#endif //#ifdef CSHOWINPUT_USECUDA
}

UINT CShowInputView::_MonitorFunc(LPVOID pParam)
{
#pragma warning(suppress: 26429)
  CShowInputView* pView{static_cast<CShowInputView*>(pParam)};
  return pView->MonitorFunc();
}

UINT CShowInputView::MonitorFunc()
{
  try
  {
    //Create an instance of the class which does the audio input
    CWaveIn reader;

    //Ask it to allocate its buffers
    reader.CreateBuffers();

    //Open the wave input device
    WAVEFORMATEX wfe{};
    wfe.wFormatTag = WAVE_FORMAT_PCM;
    wfe.nChannels = 1;
    wfe.nSamplesPerSec = 44100;
    wfe.wBitsPerSample = 16;
#pragma warning(suppress: 26472)
    wfe.nBlockAlign = static_cast<WORD>((wfe.nChannels * wfe.wBitsPerSample) >> 3);
    wfe.nAvgBytesPerSec = wfe.nSamplesPerSec * wfe.nBlockAlign;
    wfe.cbSize = 0;
    reader.Open(&wfe);

    //Now that everthing is ready to go, start input on the wave in device
    reader.Start();

    //the main record loop, basically read from the input,
    //perform an FFT of the data and then show it graphically
    while (!m_bWantStop)
    {
      //Read from the file in chunks
#pragma warning(suppress: 26485)
      reader.Read(m_inbuffer, sizeof(m_inbuffer));

      //Perform the FFT on the input data
    #ifdef CSHOWINPUT_USECUDA
      for (int i{0}; i<_countof(m_inbuffer); i++)
#pragma warning(suppress: 26446 26482)
        m_InHost[i].x = m_inbuffer[i] / static_cast<float>(_I16_MAX + 0.0);
#pragma warning(suppress: 26485)
      cudaError e{cudaMemcpy(m_InDevice, m_InHost, sizeof(cufftComplex) * _countof(m_inbuffer), ::cudaMemcpyHostToDevice)};
      VERIFY(e == cudaSuccess);
      const cufftResult r{cufftExecC2C(m_FFTPlan, m_InDevice, m_OutDevice, CUFFT_FORWARD)};
      VERIFY(r == CUFFT_SUCCESS);
      e = cudaDeviceSynchronize();
      VERIFY(e == cudaSuccess);
#pragma warning(suppress: 26485)
      e = cudaMemcpy(m_OutHost, m_OutDevice, sizeof(cufftComplex)*_countof(m_inbuffer), ::cudaMemcpyDeviceToHost);
      VERIFY(e == cudaSuccess);
    #else
      for (size_t i{0}; i<_countof(m_inbuffer); i++)
#pragma warning(suppress: 26446 26482)
        m_RealIn[i] = m_inbuffer[i] / (_I16_MAX + 0.0);
#pragma warning(suppress: 26485)
      fft_double(1024, false, m_RealIn, nullptr, m_RealOut, m_ImagOut);
    #endif //CSHOWINPUT_USECUDA

    #ifndef CSHOWINPUT_NOD2D
      Invalidate(FALSE);
      UpdateWindow();
    #else
      //Draw the FFT results
      CDC* pDC{GetDC()};
      if (pDC != nullptr)
      {
        CRect rClient;
        GetClientRect(&rClient);
        pDC->FillSolidRect(rClient, RGB(255, 255, 255));
#pragma warning(suppress: 26486)
        DrawBars(pDC, rClient);
        ReleaseDC(pDC);
      }
    #endif //#ifndef CSHOWINPUT_NOD2D
    }
  }
#pragma warning(suppress: 26496)
  catch(CWaveInException& e)
  {
  #ifndef _DEBUG
    UNREFERENCED_PARAMETER(e);
  #endif //#ifndef _DEBUG
    ATLTRACE(_T("An unexpected waveIn exception occurred, Error:%u\n"), e.m_Error);
  }
#pragma warning(suppress: 26496)
  catch(CWaveOutException& e)
  {
  #ifndef _DEBUG
    UNREFERENCED_PARAMETER(e);
  #endif //#ifndef _DEBUG
    ATLTRACE(_T("An unexpected waveOut exception occurred, Error:%u\n"), e.m_Error);
  }

  return 0;
}

BOOL CShowInputView::PreCreateWindow(CREATESTRUCT& cs)
{
  //Let the base class do its thing
  return __super::PreCreateWindow(cs);
}

#pragma warning(suppress: 26429)
void CShowInputView::OnDraw(CDC* pDC)
{
  CRect rClient;
  GetClientRect(&rClient);
  pDC->FillSolidRect(rClient, RGB(255, 255, 255));
  DrawBars(pDC, rClient);
}

#pragma warning(suppress: 26434)
CShowInputDoc* CShowInputView::GetDocument() noexcept
{
#pragma warning(suppress: 26466)
  return static_cast<CShowInputDoc*>(m_pDocument);
}

void CShowInputView::OnInitialUpdate()
{
  //Let the base class do its thing
  __super::OnInitialUpdate();

  if (m_pThread == nullptr)
  {
    m_pThread = AfxBeginThread(_MonitorFunc, this, THREAD_PRIORITY_NORMAL,CREATE_SUSPENDED);
    ASSERT(m_pThread);
    m_pThread->m_bAutoDelete = FALSE;
    m_pThread->ResumeThread();
  }
}

#pragma warning(suppress: 26434)
void CShowInputView::OnDestroy()
{
  //Wait for the worker thread to exit
  if (m_pThread != nullptr)
  {
    m_bWantStop = TRUE;
    WaitForSingleObject(m_pThread->m_hThread, INFINITE);
    delete m_pThread;
    m_pThread = nullptr;
  }

  //Let the base class do its thing
  __super::OnDestroy();
}

#pragma warning(suppress: 26429)
void CShowInputView::DrawBars(CDC* pDC, const CRect& rClient)
{
  double fMaxPower{-1};
#ifdef CSHOWINPUT_USECUDA
  for (int i{0}; i<(_countof(m_OutHost) / 2); i++)
  {
#pragma warning(suppress: 26446 26482)
    m_Power[i] = 20 * log(sqrt(m_OutHost[i].x * static_cast<double>(m_OutHost[i].x) + m_OutHost[i].y * static_cast<double>(m_OutHost[i].y)));
#pragma warning(suppress: 26446 26482)
    if (m_Power[i] > fMaxPower)
#pragma warning(suppress: 26446)
      fMaxPower = m_Power[i];
  }

  //Draw the spectrum
  const double fMaxHeight{rClient.Height()};
  int nXOffset{0};
  for (int i{0}; i<(_countof(m_OutHost) / 2); i++)
  {
#pragma warning(suppress: 26446 26482)
    pDC->Rectangle(nXOffset, rClient.bottom - static_cast<int>(m_Power[i] / fMaxPower * fMaxHeight), nXOffset + 8, rClient.bottom);
    nXOffset += 8;
  }
#else
  for (size_t i{0}; i<(_countof(m_RealOut) / 2); i++)
  {
#pragma warning(suppress: 26446 26482)
    m_Power[i] = 20 * log(sqrt(m_RealOut[i] * m_RealOut[i] + m_ImagOut[i] * m_ImagOut[i]));
#pragma warning(suppress: 26446 26482)
    if (m_Power[i] > fMaxPower)
#pragma warning(suppress: 26446 26482)
      fMaxPower = m_Power[i];
  }

  //Draw the spectrum
  const auto fMaxHeight{static_cast<double>(rClient.Height())};
  int nXOffset{0};
  for (size_t i{0}; i<(_countof(m_RealOut) / 2); i++)
  {
#pragma warning(suppress: 26446 26482)
    pDC->Rectangle(nXOffset, rClient.bottom - static_cast<int>(m_Power[i] / fMaxPower * fMaxHeight), nXOffset + 8, rClient.bottom);
    nXOffset += 8;
  }
#endif //#ifdef CSHOWINPUT_USECUDA
}

#ifndef CSHOWINPUT_NOD2D
LRESULT CShowInputView::OnDraw2D(WPARAM /*wParam*/, LPARAM lParam)
{
  //Pull out the render target
#pragma warning(suppress: 26490)
  CHwndRenderTarget* pRenderTarget{reinterpret_cast<CHwndRenderTarget*>(lParam)};
#pragma warning(suppress: 26486)
  ASSERT_VALID(pRenderTarget);

  //Call the helper function to do the heavy lifting
#pragma warning(suppress: 26486)
  Draw(pRenderTarget);

  return TRUE;
}

#pragma warning(suppress: 26440)
LRESULT CShowInputView::OnRecreatedResources(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
  return 0L;
}

#pragma warning(suppress: 26429)
void CShowInputView::DrawBars(CRenderTarget* pRenderTarget, const CRect& rClient)
{
  //Create the brush we need
  CD2DSolidColorBrush brush{pRenderTarget, D2D1::ColorF::Black};

  double fMaxPower{-1};
#ifdef CSHOWINPUT_USECUDA
  for (int i{0}; i<(_countof(m_OutHost) / 2); i++)
  {
    m_Power[i] = 20*log(sqrt(m_OutHost[i].x * m_OutHost[i].x + m_OutHost[i].y * m_OutHost[i].y));
    if (m_Power[i] > fMaxPower)
      fMaxPower = m_Power[i];
  }
  //Draw the spectrum
  double fMaxHeight{rClient.Height()};
  int nXOffset{0};
  for (int i{0}; i<(_countof(m_OutHost)/2); i++)
  {
    CD2DRectF rBar{static_cast<FLOAT>(nXOffset), static_cast<FLOAT>(rClient.bottom - (m_Power[i] / fMaxPower * fMaxHeight)), static_cast<FLOAT>(nXOffset + 8), static_cast<FLOAT>(rClient.bottom)};
    pRenderTarget->DrawRectangle(rBar, &brush);
    nXOffset += 8;
  }
#else
  for (size_t i{0}; i<(_countof(m_RealOut) / 2); i++)
  {
#pragma warning(suppress: 26446 26482)
    m_Power[i] = 20 * log(sqrt(m_RealOut[i] * m_RealOut[i] + m_ImagOut[i] * m_ImagOut[i]));
#pragma warning(suppress: 26446 26482)
    if (m_Power[i] > fMaxPower)
#pragma warning(suppress: 26446 26482)
      fMaxPower = m_Power[i];
  }

  //Draw the spectrum
  const auto fMaxHeight{static_cast<double>(rClient.Height())};
  int nXOffset{0};
  for (size_t i{0}; i<(_countof(m_RealOut) / 2); i++)
  {
#pragma warning(suppress: 26446 26482)
    const CD2DRectF rBar{static_cast<FLOAT>(nXOffset), static_cast<FLOAT>(rClient.bottom - (m_Power[i] / fMaxPower * fMaxHeight)), static_cast<FLOAT>(nXOffset + 8), static_cast<FLOAT>(rClient.bottom)};
    pRenderTarget->DrawRectangle(rBar, &brush);
    nXOffset += 8;
  }
#endif //#ifdef CSHOWINPUT_USECUDA
}

HRESULT CShowInputView::Draw(_In_ CRenderTarget* pRenderTarget)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(pRenderTarget != nullptr);

  CRect rClient;
  GetClientRect(&rClient);
  const CD2DRectF d2dClient{rClient};

  //Create the brush we need
  CD2DSolidColorBrush brush{pRenderTarget, D2D1::ColorF::White};

  pRenderTarget->FillRectangle(d2dClient, &brush);
  DrawBars(pRenderTarget, rClient);

  return S_OK;
}
#endif //ifndef CSHOWINPUT_NOD2D

#pragma warning(suppress: 26434 26440)
BOOL CShowInputView::OnEraseBkgnd(CDC* /*pDC*/)
{
  //Nothing to do
  return TRUE;
}
